"""Implementation of GBA stoic Problem class.

based on gba_problem
- extended Growth Balance Analysis (GBA) problem from Deniz Sezer
- basic problem uses internal metabolite and enzyme concentrations
  as free variables (optimization variables)
  it optimizes growth rate under constraints of mass balance and
  total density (cellular density or macromolecular density)
- extended formulation takes care about
  - protein degradation
  - protein interconversion
  - spontaneous reactions

- a related formulation is based on Hugo Dourado scalled fluxes
  see gba_sflux_problem.py

- this implementation uses the JAX package to
  - determine first and second derivatives of all functions
  - just in time compilation for speed-up
  - support of accellerators (e.g. GPUs) - yet to be tested

Peter Schubert, HHU Duesseldorf, September 2022
"""

import numpy as np
import re
import types
import scipy
import jax
import jax.numpy as jnp

jax.config.update("jax_enable_x64", True)

# TODO: variable transformation x = exp(y) and solve for unbounded y. start with vector y = log(x)


class GbaStoicProblem:

    cache = True
    last_values = {}

    def __init__(self, xba_model):
        """Initialize GbaStoichProblem from a xba_model.

        :param xba_model: xba model with all required GBA parameters
        :type xba_model: XbaModel
        """
        self.xba_model = xba_model

        # check to scale density constraint to improve convergence
        self.scale_density = 1e-4

        # metabolites and enzymes in the model
        metab_sids = [s.id for s in xba_model.species.values()
                      if s.sboterm.is_in_branch('SBO:0000247')]

        # optimization variables
        self.var_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is False]
        self.var_enz_sids = [s.id for s in xba_model.species.values()
                             if s.sboterm.is_in_branch('SBO:0000246')]
        self.var_sids = self.var_metab_sids + self.var_enz_sids
        self.n_vars = len(self.var_sids)
        self.var_sid2idx = {sid: idx for idx, sid in enumerate(self.var_sids)}
        self.var_mask_metab = [True if sid in self.var_metab_sids else False for sid in self.var_sids]
        self.var_mask_enz = [True if sid in self.var_enz_sids else False for sid in self.var_sids]
        self.var_initial = np.array([xba_model.species[sid].initial_conc for sid in self.var_sids])
        self.var_mws = np.array([xba_model.species[sid].mw for sid in self.var_sids])
        self.var_vols = np.array([xba_model.compartments[xba_model.species[sid].compartment].size
                                  for sid in self.var_sids])

        # constant concentrations
        self.const_metab_sids = [sid for sid in metab_sids if xba_model.species[sid].constant is True]
        self.const_sid2idx = {sid: idx for idx, sid in enumerate(self.const_metab_sids)}

        # protein synthesis reactions get ordered according to the order of their enzyme products
        #  to facilitate matrix multiplication without further remapping
        # - here we expect a one-to-one relationship between enzyme and corresponding protein synthesis reaction
        self.mr_rids = [r.id for r in xba_model.reactions.values()
                        if r.sboterm.is_in_branch('SBO:0000167')]
        self.ps_rids = [xba_model.species[sid].ps_rid for sid in self.var_enz_sids]
        self.mr_vols = np.array([xba_model.compartments[xba_model.reactions[rid].compartment].size
                                 for rid in self.mr_rids])
        self.ps_vols = np.array([xba_model.compartments[xba_model.reactions[rid].compartment].size
                                 for rid in self.ps_rids])

        # calculate stoichiometric sub-matrices and scale them according to volume
        # TODO: sparce matrix storage and calculations
        species_x_mrs = xba_model.get_stoic_matrix(self.var_sids, self.mr_rids)
        self.dof = species_x_mrs.shape[1] - np.linalg.matrix_rank(species_x_mrs)
        metab_x_mrs = xba_model.get_stoic_matrix(self.var_metab_sids, self.mr_rids)
        self.metab_x_psrs = xba_model.get_stoic_matrix(self.var_metab_sids, self.ps_rids)
        self.enz_x_mrs = xba_model.get_stoic_matrix(self.var_enz_sids, self.mr_rids)
        self.enz_vols = self.var_vols[self.var_mask_enz]
        metab_vols = self.var_vols[self.var_mask_metab]

        self.metab_x_mras = metab_x_mrs / metab_vols.reshape((-1, 1))
        self.metab_x_psras = self.metab_x_psrs / metab_vols.reshape((-1, 1))
        self.enz_x_mras = self.enz_x_mrs / self.enz_vols.reshape((-1, 1))

        # identify reactions ids for enzyme degradation and corresponding submatrix
        enz_degrad_idxs = self.get_export_idxs(self.enz_x_mrs)
        enz_degrad_rids = np.array(self.mr_rids)[enz_degrad_idxs]
        self.enz_x_degrads = self.enz_x_mrs[:, enz_degrad_idxs]

        # identify reactions ids for enzyme transitions and corresponding submatrix
        # for submatrix perform LU decomposition and remove empty rows from U matrix
        enz_trans_idxs = self.get_trans_idxs(self.enz_x_mras)
        enz_trans_rids = np.array(self.mr_rids)[enz_trans_idxs]
        u = scipy.linalg.lu((self.enz_x_mras[:, enz_trans_idxs]))[2]
        self.subenz_x_trans = u[abs(u).sum(axis=1) > 0]

        # model parameters
        self.model_params = {pid: p.value for pid, p in xba_model.parameters.items()}
        self.model_params.update({cid: c.size for cid, c in xba_model.compartments.items()})
        self.model_params['macro'] = self.model_params.get('macro', 0.0) > 0.0
        self.model_params['ext_conc'] = np.array([xba_model.species[sid].initial_conc
                                                  for sid in self.const_metab_sids])

        # move some numpy structures to jax.numpy
        # TODO: possibly code relevant numpy objects in jax.numpy to allow for accelleration
        self.jvar_mask_enz = jnp.array(self.var_mask_enz)
        self.jenz_x_mrs = jnp.array(self.enz_x_mrs)
        self.jvar_mask_metab = jnp.array(self.var_mask_metab)
        self.jenz_x_degrads = jnp.array(self.enz_x_degrads)
        self.jenz_vols = jnp.array(self.enz_vols)
        self.jsubenz_x_trans = jnp.array(self.subenz_x_trans)
        self.jvar_mws = jnp.array(self.var_mws)
        self.jmetab_x_mras = jnp.array(self.metab_x_mras)
        self.jmetab_x_psrs = jnp.array(self.metab_x_psrs)
        self.jenz_x_mras = jnp.array(self.enz_x_mras)
        self.jmetab_x_psrs = jnp.array(self.metab_x_psrs)

        # retrieve kinetic functions and compile them
        self.mras = self.get_reactions(self.mr_rids, inv=False)
        self.pstmas = self.get_reactions(self.ps_rids, inv=True)
        self.jmras_jit = jax.jit(self.mras)
        self.jpstmas_jit = jax.jit(self.pstmas)
        self.jdegradas_jit = jax.jit(self.get_reactions(enz_degrad_rids, inv=False))
        self.jenz_tras_jit = jax.jit(self.get_reactions(enz_trans_rids, inv=False))

        # create functions used for optimization
        self.jobj_mu_jit = jax.jit(self.get_growth_rate)
        self.jobj_mu_grad_jit = jax.jit(jax.grad(self.jobj_mu_jit))
        self.jobj_mu_hess_jit = jax.jit(jax.jacfwd(self.jobj_mu_grad_jit))

        self.jeq_density_jit = jax.jit(self.heq_density)
        self.jeq_density_jac_jit = jax.jit(jax.jacrev(self.jeq_density_jit))
        self.jeq_density_hesss_jit = jax.jit(jax.jacfwd(self.jeq_density_jac_jit))
        self.jeq_density_macro_jit = jax.jit(self.heq_density_macro)
        self.jeq_density_macro_jac_jit = jax.jit(jax.jacrev(self.jeq_density_macro_jit))
        self.jeq_density_macro_hesss_jit = jax.jit(jax.jacfwd(self.jeq_density_macro_jac_jit))

        self.jeq_mass_balance_jit = jax.jit(self.heq_mass_balance)
        self.jeq_mass_balance_jac_jit = jax.jit(jax.jacrev(self.jeq_mass_balance_jit))
        self.jeq_mass_balance_hesss_jit = jax.jit(jax.jacfwd(self.jeq_mass_balance_jac_jit))

        self.jeq_enz_trans_jit = jax.jit(self.heq_enz_trans)
        self.jeq_enz_trans_jac_jit = jax.jit(jax.jacrev(self.jeq_enz_trans_jit))
        self.jeq_enz_trans_hesss_jit = jax.jit(jax.jacfwd(self.jeq_enz_trans_jac_jit))

        self.n_constraints = {'heq_mass_balance': len(self.var_metab_sids),
                              'heq_enz_trans': self.subenz_x_trans.shape[0],
                              'heq_density': 1}

    def compile(self, x, model_params):
        jx = jnp.array(x)

        self.jmras_jit(jx, model_params)
        self.jpstmas_jit(jx, model_params)
        self.jdegradas_jit(jx, model_params)
        self.jenz_tras_jit(jx, model_params)

        self.jobj_mu_jit(jx, model_params)
        self.jobj_mu_grad_jit(jx, model_params)
        self.jobj_mu_hess_jit(jx, model_params)

        self.jeq_density_jit(jx, model_params)
        self.jeq_density_jac_jit(jx, model_params)
        self.jeq_density_hesss_jit(jx, model_params)
        self.jeq_density_macro_jit(jx, model_params)
        self.jeq_density_macro_jac_jit(jx, model_params)
        self.jeq_density_macro_hesss_jit(jx, model_params)

        self.jeq_mass_balance_jit(jx, model_params)
        self.jeq_mass_balance_jac_jit(jx, model_params)
        self.jeq_mass_balance_hesss_jit(jx, model_params)

        self.jeq_enz_trans_jit(jx, model_params)
        self.jeq_enz_trans_jac_jit(jx, model_params)
        self.jeq_enz_trans_hesss_jit(jx, model_params)

    def update_model_params(self, key, val):
        self.model_params[key] = val

    def _to_vectors(self, rs_str):
        """replace variables and constant metabolites by vector elements.

        Variables converted to x[], external metabolites to ext_conc[].
        access to model parameters via model_params dict

        :param rs_str: reaction strings (extracted from the xba model).
        :type rs_str: list of strings
        :returns: rs_v_str
        :rtype: list of reaction stings with parameters replaced by vector elements
        """
        rs_v_str = []
        for r_v_str in rs_str:
            for idx, sid in enumerate(self.var_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'x[{idx}]', r_v_str)
            for idx, sid in enumerate(self.const_metab_sids):
                r_v_str = re.sub(r'\b' + sid + r'\b', f'ext_conc[{idx}]', r_v_str)
            for param_key in self.model_params:
                r_v_str = re.sub(r'\b' + param_key + r'\b', f'model_params["{param_key}"]', r_v_str)
            rs_v_str.append(r_v_str)
        return rs_v_str

    def get_reactions(self, rids, inv=False):
        """create function objects for reactions rates/times within jax.numpy namespace

        extracted from kinetic laws for SBML model

        Note: reactions in the SBML model are in amounts per time (not concentration per time)
        I.e. for mass balances equations, we have to devide by volumes where metabolite is located

        :param rids: reaction ids for reactions in amount per time to be processed.
        :type rids: list of strings
        :param inv: flag indicating if inverse of reaction required (i.e. reaction times)
        :type inv: boolean (default: False)
        :returns: jras
        :rtype: function object, returning an 1D array of function results
        """
        if inv is False:
            kinetics_str = [self.xba_model.reactions[rid].expanded_kl for rid in rids]
        else:
            kinetics_str = [f'1.0/({self.xba_model.reactions[rid].expanded_kl})' for rid in rids]
        v_str = self._to_vectors(kinetics_str)
        func_code = compile('def reactions(x, model_params): '
                            'return np.array([' + ', '.join(v_str) + '])',
                            '<string>', 'exec')
        jras = types.FunctionType(func_code.co_consts[0], {'np': jnp})
        return jras

    @staticmethod
    def get_trans_idxs(smat):
        """Identify metabolic reactions in S matrix with species transitions.

        I.e. columns containing both positive and negative entries

        :param smat:
        :type smat:
        :return: bolean vector identifying relevant columns
        :rtype: 1D ndarray with dtype bool
        """
        mask_trans_idxs = np.zeros(smat.shape[1])
        for col_idx in range(smat.shape[1]):
            if ((np.count_nonzero(smat[:, col_idx] < 0) > 0) and
                    np.count_nonzero(smat[:, col_idx] > 0) > 0):
                mask_trans_idxs[col_idx] = 1.0
        return mask_trans_idxs > 0

    @staticmethod
    def get_export_idxs(smat):
        """Identify metabolic reactions in S matrix with consumption only.

        I.e. columns containing only negative entries

        :param smat:
        :type smat:
        :return: bolean vector identifying relevant columns
        :rtype: 1D ndarray with dtype bool
        """
        mask_remove_idxs = np.zeros(smat.shape[1])
        for col_idx in range(smat.shape[1]):
            if ((np.count_nonzero(smat[:, col_idx] < 0) > 0) and
                    np.count_nonzero(smat[:, col_idx] > 0) == 0):
                mask_remove_idxs[col_idx] = 1.0
        return mask_remove_idxs > 0

    # @jax.jit
    def get_growth_rate_0(self, x, model_params):
        """Calculate growth rate at x, assuming no protein degradation

        Equations:
            gr0(x) = 1/tau(x)
            tau(x) = pstms(x).dot(ce)
            pstms(x) = enz_vols * ptsmas(x)

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: growth rate at x (without protein degradation)
        :rtype: float
        """
        jx = jnp.array(x)
        ce = jx[self.jvar_mask_enz]
        pstms = self.jenz_vols * self.jpstmas_jit(jx, model_params)
        gr0 = jnp.divide(1.0, jnp.dot(pstms, ce))
        return gr0

    # @jax.jit
    def get_growth_rate(self, x, model_params):
        """Calculate growth rate at x, considering protein degradation

        considering stoichiometric coefficiens unequal unity for enzyme turnover
        stoichiometric sub-matrices do not require volume scaling (as Vol_enz gets factored out)
        With degradation (!!negative stoichiometric coefficients in enz_x_mrs), growth rate reduces
        - first calculate the enzyme amount per time being degraded

        Equations:
            gr(x) = gr0(x) * (1 + enz_to_mras(x).dot(pstmas(x)))
            enz_to_mras(x) = enz_x_mrs.dot(mras(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: growth rate at x (considering protein degradation)
        :rtype: float
        """
        jx = jnp.array(x)
        gr0 = self.get_growth_rate_0(jx, model_params)
        pstmas = self.jpstmas_jit(jx, model_params)
        enz_to_mras = jnp.dot(self.jenz_x_degrads, self.jdegradas_jit(x, model_params))
        gr = gr0 * (1.0 + jnp.dot(enz_to_mras, pstmas))
        return gr

    def get_alphas(self, x, model_params):
        """Calculate ribosomal allocations at x, considering protein degradation

        alpha_e = gr * ce_e * pstm_e + p_degr_e * pstm_e
        alpha = gr * ce.dot(pstms(x)) + p_degrs(x).dot(pstms(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: ribosomal allocations at x (considering protein degradation)
        :rtype: numpy.ndarray (1D)
        """
        jx = jnp.array(x)
        ce = jx[self.jvar_mask_enz]
        gr = self.get_growth_rate(jx, model_params)
        pstms = self.jenz_vols * self.jpstmas_jit(jx, model_params)

        alphas0 = gr * ce * pstms
        pstmas = self.jpstmas_jit(jx, model_params)
        enz_to_mras = jnp.dot(self.jenz_x_degrads, self.jdegradas_jit(jx, model_params))
        alphas = alphas0 - enz_to_mras * pstmas
        return alphas

    def heq_mass_balance(self, x, model_params):
        """Calculate mass balance equations at x for metabolites

        incuding protein degradation, which consume metabolites
        Note: protein degradation is also part of metabolic reactions (mras), which return
              metabolites.

        metab_mb = dcm/dt = metabolic_balance + growth_dilution + sythesis_enzyme_balance
          metabolic_balance = metab_x_mras.dot(mras(x))
          growth_dilution = λ * (metab_x_psrs.dot(ce) - cm)
            ce_scaled = enz_vols * ce   # convert concentration to amounts, considering stoichiometry
          sythesis_enzyme_balance = - metab_x_psras.dot(enz_vols * enzyme_balance)
            enzyme_balance = enz_x_mras.dot(mras(x))

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: mass balance calculated at x
        :rtype: numpy.ndarray (1D) with shape (len(var_mask_metab), 1)
        """
        jx = jnp.array(x)
        gr = self.get_growth_rate(jx, model_params)
        cm = jx[self.jvar_mask_metab]
        ce = jx[self.jvar_mask_enz]

        metabolic_mb = jnp.dot(self.jmetab_x_mras, self.jmras_jit(jx, model_params))
        dilution = jnp.dot(self.jmetab_x_psrs, ce) - cm
        metab_mb = metabolic_mb + gr * dilution
        enz_mb = jnp.dot(self.jenz_x_mras, self.jmras_jit(jx, model_params))
        metab_mb -= jnp.dot(self.jmetab_x_psrs, enz_mb)
        return metab_mb

    def heq_enz_trans(self, x, model_params):
        """Calculate enzyme transition equations at x

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: Jacobian of mass balance equations at x
        :rtype: numpy.ndarray (1D) with shape (len(subenz_x_trans.shape[1]), 1)
        """
        jx = jnp.array(x)
        return jnp.dot(self.jsubenz_x_trans, self.jenz_tras_jit(jx, model_params))

    def heq_density(self, x, model_params):
        """Calculate dry mass density (g/l) constraint at x

        Could be extended having several density constraints (not yet implemented)

        alternalively, we could check macromolecular (protein) density using heq_density_macro

        using ipopt, the constraint bounds could be set to rho (only add up masses)
        density = sum(species_concentration * molecular_weight)

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: mass balance calculated at x
        :rtype: numpy.ndarray (1D) with shape (1,) for now
        """
        jx = jnp.array(x)
        density = jnp.dot(self.jvar_mws, jx)
        return jnp.array([model_params['rho'] - density]) * self.scale_density

    def heq_density_macro(self, x, model_params):
        """Calculate macromolecular density (g/l) constraint at x

        Could be extended having several density constraints (not yet implemented)

        macromolecular density is total protein mass density

        using ipopt, the constraint bounds could be set to rho (only add up masses)
        density = sum(species_concentration * molecular_weight)

        :param x: optimization variables
        :type x: numpy.ndarray
        :param model_params: model parameters that can be changed
        :type model_params: dict
        :returns: mass balance calculated at x
        :rtype: numpy.ndarray (1D) with shape (1,) for now
        """
        jx = jnp.array(x)
        ce = jx[self.jvar_mask_enz]
        density = jnp.dot(self.jvar_mws[self.jvar_mask_enz], ce)
        return jnp.array([model_params['rho'] - density]) * self.scale_density
