"""Implementation of IpoptGbaProblem class.

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np


# based on cyipopt examples
class IpoptGbaProblem:

    def __init__(self, gba_problem):
        """Initialize GbaIpoptProblem.

        :param gba_problem: GbaProblem from where to extract relevant parameters
        :type gba_problem: GbaProblem
        """
        self.gba_problem = gba_problem
        self._n_vars = len(gba_problem.var_initial)
        self.report_freq = 0

        self.nfev = 0
        self.njev = 0
        self.nit = 0
        self.constraint_dims = sum(gba_problem.n_constraints.values())

    def objective(self, x):
        self.nfev += 1
        return -self.gba_problem.get_growth_rate(x)

    def gradient(self, x):
        self.njev += 1
        return -self.gba_problem.get_growth_rate_grad(x)

    def constraints(self, x):
        # TODO: improve code
        if self.gba_problem.n_constraints['heq_enz_trans'] > 0:
            eqs = np.hstack((self.gba_problem.heq_mass_balance(x),
                             self.gba_problem.heq_enz_trans(x),
                             self.gba_problem.heq_density(x)))
        else:
            eqs = np.hstack((self.gba_problem.heq_mass_balance(x),
                             self.gba_problem.heq_density(x)))
        return eqs

    def jacobian(self, x):
        # TODO: improve code
        if self.gba_problem.n_constraints['heq_enz_trans'] > 0:
            jacs = (np.vstack((self.gba_problem.heq_mass_balance_jac(x),
                               self.gba_problem.heq_enz_trans_jac(x),
                               self.gba_problem.heq_density_jac(x)))
                    ).flatten()
        else:
            jacs = (np.vstack((self.gba_problem.heq_mass_balance_jac(x),
                               self.gba_problem.heq_density_jac(x)))
                    ).flatten()
        return jacs

    # def hessianstructure(self):
    #   return np.tril_indices(self._n_vars)

    def hessian(self, x, lagrange, obj_factor):
        hess = obj_factor * -self.gba_problem.get_growth_rate_hess(x)
        if self.gba_problem.n_constraints['heq_enz_trans'] > 0:
            hesss = np.vstack((self.gba_problem.heq_mass_balance_hess(x),
                               self.gba_problem.heq_enz_trans_hess(x),
                               self.gba_problem.heq_density_hess(x)))
        else:
            hesss = np.vstack((self.gba_problem.heq_mass_balance_hess(x),
                               self.gba_problem.heq_density_hess(x)))
        hess += (lagrange.reshape((-1, 1, 1)) * hesss).sum(axis=0)
        return hess[np.tril_indices(self._n_vars)]

    def set_report_freq(self, mod_val=0):
        self.report_freq = mod_val

    # noinspection PyUnusedLocal
    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr,
                     ls_trials):
        self.nit = iter_count
        if self.report_freq > 0:
            if iter_count % self.report_freq == 0:
                print(f'[{iter_count:5d}] growth rate: {-3600 *obj_value:.5f}, '
                      f"density constr: {self.gba_problem.last_values['heq_density'][1]}")
