"""Implementation of IpoptGbaSfluxProblem class.

For solving GBA problem using scaled fluxes as per Hugo Dourado

Peter Schubert, HHU Duesseldorf, September 2022
"""

import numpy as np
import jax.numpy as jnp


# based on cyipopt examples
class IpoptGbaSfluxProblem:

    def __init__(self, gba_sflux_problem, model_params=None):
        """Initialize sFluxIpoptProblem.

        :param gba_sflux_problem: GbaSfluxProblem from where to extract relevant parameters
        :type gba_sflux_problem: GbaSfluxProblem
        :param model_params: model parameters (optional, default None, used params from GbaSfluxProblem
        :type model_params: None or dict
        """
        self.sflux_problem = gba_sflux_problem
        if model_params is None:
            self.model_params = gba_sflux_problem.model_params
        else:
            self.model_params = model_params
        self._n_vars = len(gba_sflux_problem.rids)
        self.report_freq = 0

        self.nfev = 0
        self.njev = 0
        self.nit = 0
        self.constraint_dims = sum(gba_sflux_problem.n_constraints.values())

    def objective(self, x):
        jx = jnp.array(x)
        self.nfev += 1
        return -np.float64(self.sflux_problem.obj_mu(jx, self.model_params))

    def gradient(self, x):
        jx = jnp.array(x)
        self.njev += 1
        return -np.array(self.sflux_problem.obj_mu_grad(jx, self.model_params))

    def constraints(self, x):
        jx = jnp.array(x)
        # TODO: improve code
        eqs = np.hstack((self.sflux_problem.constr_density(jx, self.model_params),
                         self.sflux_problem.constrs_conc(jx, self.model_params)))
        return eqs

    def jacobian(self, x):
        # TODO: improve code
        jx = jnp.array(x)
        jacs = (np.vstack((self.sflux_problem.constr_density_grad(jx, self.model_params),
                           self.sflux_problem.constrs_conc_jac(jx, self.model_params)))
                ).flatten()
        return jacs

    # def hessianstructure(self):
    #    return np.tril_indices(self._n_vars)

    def hessian(self, x, lagrange, obj_factor):
        jx = jnp.array(x)
        hess = obj_factor * -self.sflux_problem.obj_mu_hess(jx, self.model_params)
        hesss = np.vstack(([self.sflux_problem.constr_density_hess(jx, self.model_params)],
                           self.sflux_problem.constrs_conc_hesss(jx, self.model_params)))
        hess += (lagrange.reshape((-1, 1, 1)) * hesss).sum(axis=0)
        return hess[np.tril_indices(self._n_vars)]

    def set_report_freq(self, mod_val=0):
        self.report_freq = mod_val

    # noinspection PyUnusedLocal
    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr, ls_trials):
        self.nit = iter_count
        if self.report_freq > 0:
            if iter_count % self.report_freq == 0:
                print(f'[{iter_count:5d}] growth rate: {-obj_value:.4f} 1/h')
