"""Implementation of IpoptGbaStoicProblem class.

For solving GBA problem using scaled fluxes as per Hugo Dourado

Peter Schubert, HHU Duesseldorf, September 2022
"""

import numpy as np
import jax.numpy as jnp


# based on cyipopt examples
class IpoptGbaStoicProblem:

    def __init__(self, gba_stoic_problem, model_params=None):
        """Initialize sFluxIpoptProblem.

        :param gba_stoic_problem: GbaStoicProblem from where to extract relevant parameters
        :type gba_stoic_problem: GbaStoicProblem
        :param model_params: model parameters (optional, default None, use params from GbaStoicProblem
        :type model_params: None or dict
        """
        self.stoic_problem = gba_stoic_problem
        if model_params is None:
            self.model_params = gba_stoic_problem.model_params
        else:
            self.model_params = model_params
        self._n_vars = len(gba_stoic_problem.var_initial)
        self.report_freq = 0

        self.nfev = 0
        self.njev = 0
        self.nit = 0
        self.constraint_dims = sum(gba_stoic_problem.n_constraints.values())

    def objective(self, x):
        jx = jnp.array(x)
        self.nfev += 1
        return -np.float64(self.stoic_problem.jobj_mu_jit(jx, self.model_params))

    def gradient(self, x):
        jx = jnp.array(x)
        self.njev += 1
        return -np.array(self.stoic_problem.jobj_mu_grad_jit(jx, self.model_params))

    def constraints(self, x):
        jx = jnp.array(x)
        if self.stoic_problem.n_constraints['heq_enz_trans'] > 0:
            eqs = np.hstack((self.stoic_problem.jeq_mass_balance_jit(jx, self.model_params),
                             self.stoic_problem.jeq_enz_trans_jit(jx, self.model_params),
                             self.stoic_problem.jeq_density_jit(jx, self.model_params)))
        else:
            eqs = np.hstack((self.stoic_problem.jeq_mass_balance_jit(jx, self.model_params),
                             self.stoic_problem.jeq_density_jit(jx, self.model_params)))
        return eqs

    def jacobian(self, x):
        jx = jnp.array(x)
        if self.stoic_problem.n_constraints['heq_enz_trans'] > 0:
            jacs = (np.vstack((self.stoic_problem.jeq_mass_balance_jac_jit(jx, self.model_params),
                               self.stoic_problem.jeq_enz_trans_jac_jit(jx, self.model_params),
                               self.stoic_problem.jeq_density_jac_jit(jx, self.model_params)))
                    ).flatten()
        else:
            jacs = (np.vstack((self.stoic_problem.jeq_mass_balance_jac_jit(jx, self.model_params),
                               self.stoic_problem.jeq_density_jac_jit(jx, self.model_params)))
                    ).flatten()
        return jacs

    # def hessianstructure(self):
    #    return np.tril_indices(self._n_vars)

    def hessian(self, x, lagrange, obj_factor):
        jx = jnp.array(x)
        hess = obj_factor * -self.stoic_problem.jobj_mu_hess_jit(x, self.model_params)
        if self.stoic_problem.n_constraints['heq_enz_trans'] > 0:
            hesss = np.vstack((self.stoic_problem.jeq_mass_balance_hesss_jit(jx, self.model_params),
                               self.stoic_problem.jeq_enz_trans_hesss_jit(jx, self.model_params),
                               self.stoic_problem.jeq_density_hesss_jit(jx, self.model_params)))
        else:
            hesss = np.vstack((self.stoic_problem.jeq_mass_balance_hesss_jit(jx, self.model_params),
                               self.stoic_problem.jeq_density_hesss_jit(jx, self.model_params)))
        hess += (lagrange.reshape((-1, 1, 1)) * hesss).sum(axis=0)
        return hess[np.tril_indices(self._n_vars)]

    def set_report_freq(self, mod_val=0):
        self.report_freq = mod_val

    # noinspection PyUnusedLocal
    def intermediate(self, alg_mod, iter_count, obj_value, inf_pr, inf_du, mu,
                     d_norm, regularization_size, alpha_du, alpha_pr, ls_trials):
        self.nit = iter_count
        if self.report_freq > 0:
            if iter_count % self.report_freq == 0:
                print(f'[{iter_count:5d}] growth rate: {-obj_value:.4f} 1/h')
